#include "xproto.h"

#include <xcb/xcb.h>
#include <xcb/xcbext.h>

#include "base/logging.h"
#include "ui/gfx/x/xproto_internal.h"

namespace x11 {

XProto::XProto(XDisplay* display) : display_(display) {}

Future<void>
XProto::CreateWindow(
    const XProto::CreateWindowRequest& request) {
  WriteBuffer buf;

  auto& depth = request.depth;
  auto& wid = request.wid;
  auto& parent = request.parent;
  auto& x = request.x;
  auto& y = request.y;
  auto& width = request.width;
  auto& height = request.height;
  auto& border_width = request.border_width;
  auto& c_class = request.c_class;
  auto& visual = request.visual;
  auto& value_mask = request.value_mask;
  auto& value_list = request.value_list;

  // major_opcode
  uint8_t major_opcode = 1;
  Write(&major_opcode, &buf);

  // depth
  Write(&depth, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // wid
  Write(&wid, &buf);

  // parent
  Write(&parent, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // border_width
  Write(&border_width, &buf);

  // c_class
  uint16_t tmp0;
  tmp0 = static_cast<uint16_t>(c_class);
  Write(&tmp0, &buf);

  // visual
  Write(&visual, &buf);

  // value_mask
  uint32_t tmp1;
  tmp1 = static_cast<uint32_t>(value_mask);
  Write(&tmp1, &buf);

  // value_list
  {
    auto& background_pixmap = value_list.background_pixmap;
    auto& background_pixel = value_list.background_pixel;
    auto& border_pixmap = value_list.border_pixmap;
    auto& border_pixel = value_list.border_pixel;
    auto& bit_gravity = value_list.bit_gravity;
    auto& win_gravity = value_list.win_gravity;
    auto& backing_store = value_list.backing_store;
    auto& backing_planes = value_list.backing_planes;
    auto& backing_pixel = value_list.backing_pixel;
    auto& override_redirect = value_list.override_redirect;
    auto& save_under = value_list.save_under;
    auto& event_mask = value_list.event_mask;
    auto& do_not_propogate_mask = value_list.do_not_propogate_mask;
    auto& colormap = value_list.colormap;
    auto& cursor = value_list.cursor;

    auto value_list_expr = value_mask;
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackPixmap)) {
      // background_pixmap
      Write(&background_pixmap, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackPixel)) {
      // background_pixel
      Write(&background_pixel, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BorderPixmap)) {
      // border_pixmap
      Write(&border_pixmap, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BorderPixel)) {
      // border_pixel
      Write(&border_pixel, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BitGravity)) {
      // bit_gravity
      uint32_t tmp2;
      tmp2 = static_cast<uint32_t>(bit_gravity);
      Write(&tmp2, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::WinGravity)) {
      // win_gravity
      uint32_t tmp3;
      tmp3 = static_cast<uint32_t>(win_gravity);
      Write(&tmp3, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackingStore)) {
      // backing_store
      uint32_t tmp4;
      tmp4 = static_cast<uint32_t>(backing_store);
      Write(&tmp4, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackingPlanes)) {
      // backing_planes
      Write(&backing_planes, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackingPixel)) {
      // backing_pixel
      Write(&backing_pixel, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::OverrideRedirect)) {
      // override_redirect
      Write(&override_redirect, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::SaveUnder)) {
      // save_under
      Write(&save_under, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::EventMask)) {
      // event_mask
      uint32_t tmp5;
      tmp5 = static_cast<uint32_t>(event_mask);
      Write(&tmp5, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::DontPropagate)) {
      // do_not_propogate_mask
      uint32_t tmp6;
      tmp6 = static_cast<uint32_t>(do_not_propogate_mask);
      Write(&tmp6, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::Colormap)) {
      // colormap
      Write(&colormap, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::Cursor)) {
      // cursor
      Write(&cursor, &buf);
    }
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ChangeWindowAttributes(
    const XProto::ChangeWindowAttributesRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;
  auto& value_mask = request.value_mask;
  auto& value_list = request.value_list;

  // major_opcode
  uint8_t major_opcode = 2;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // value_mask
  uint32_t tmp7;
  tmp7 = static_cast<uint32_t>(value_mask);
  Write(&tmp7, &buf);

  // value_list
  {
    auto& background_pixmap = value_list.background_pixmap;
    auto& background_pixel = value_list.background_pixel;
    auto& border_pixmap = value_list.border_pixmap;
    auto& border_pixel = value_list.border_pixel;
    auto& bit_gravity = value_list.bit_gravity;
    auto& win_gravity = value_list.win_gravity;
    auto& backing_store = value_list.backing_store;
    auto& backing_planes = value_list.backing_planes;
    auto& backing_pixel = value_list.backing_pixel;
    auto& override_redirect = value_list.override_redirect;
    auto& save_under = value_list.save_under;
    auto& event_mask = value_list.event_mask;
    auto& do_not_propogate_mask = value_list.do_not_propogate_mask;
    auto& colormap = value_list.colormap;
    auto& cursor = value_list.cursor;

    auto value_list_expr = value_mask;
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackPixmap)) {
      // background_pixmap
      Write(&background_pixmap, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackPixel)) {
      // background_pixel
      Write(&background_pixel, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BorderPixmap)) {
      // border_pixmap
      Write(&border_pixmap, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BorderPixel)) {
      // border_pixel
      Write(&border_pixel, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BitGravity)) {
      // bit_gravity
      uint32_t tmp8;
      tmp8 = static_cast<uint32_t>(bit_gravity);
      Write(&tmp8, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::WinGravity)) {
      // win_gravity
      uint32_t tmp9;
      tmp9 = static_cast<uint32_t>(win_gravity);
      Write(&tmp9, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackingStore)) {
      // backing_store
      uint32_t tmp10;
      tmp10 = static_cast<uint32_t>(backing_store);
      Write(&tmp10, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackingPlanes)) {
      // backing_planes
      Write(&backing_planes, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::BackingPixel)) {
      // backing_pixel
      Write(&backing_pixel, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::OverrideRedirect)) {
      // override_redirect
      Write(&override_redirect, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::SaveUnder)) {
      // save_under
      Write(&save_under, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::EventMask)) {
      // event_mask
      uint32_t tmp11;
      tmp11 = static_cast<uint32_t>(event_mask);
      Write(&tmp11, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::DontPropagate)) {
      // do_not_propogate_mask
      uint32_t tmp12;
      tmp12 = static_cast<uint32_t>(do_not_propogate_mask);
      Write(&tmp12, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::Colormap)) {
      // colormap
      Write(&colormap, &buf);
    }
    if (BitAnd(value_list_expr, CreateWindowAttribute::Cursor)) {
      // cursor
      Write(&cursor, &buf);
    }
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetWindowAttributesReply>
XProto::GetWindowAttributes(
    const XProto::GetWindowAttributesRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 3;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetWindowAttributesReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetWindowAttributesReply>
detail::ReadReply<XProto::GetWindowAttributesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetWindowAttributesReply>();

  auto& backing_store = (*reply).backing_store;
  auto& visual = (*reply).visual;
  auto& c_class = (*reply).c_class;
  auto& bit_gravity = (*reply).bit_gravity;
  auto& win_gravity = (*reply).win_gravity;
  auto& backing_planes = (*reply).backing_planes;
  auto& backing_pixel = (*reply).backing_pixel;
  auto& save_under = (*reply).save_under;
  auto& map_is_installed = (*reply).map_is_installed;
  auto& map_state = (*reply).map_state;
  auto& override_redirect = (*reply).override_redirect;
  auto& colormap = (*reply).colormap;
  auto& all_event_masks = (*reply).all_event_masks;
  auto& your_event_mask = (*reply).your_event_mask;
  auto& do_not_propagate_mask = (*reply).do_not_propagate_mask;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // backing_store
  uint8_t tmp13;
  Read(&tmp13, &buf);
  backing_store = static_cast<XProto::BackingStore>(tmp13);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // visual
  Read(&visual, &buf);

  // c_class
  uint16_t tmp14;
  Read(&tmp14, &buf);
  c_class = static_cast<XProto::WindowClass>(tmp14);

  // bit_gravity
  uint8_t tmp15;
  Read(&tmp15, &buf);
  bit_gravity = static_cast<XProto::Gravity>(tmp15);

  // win_gravity
  uint8_t tmp16;
  Read(&tmp16, &buf);
  win_gravity = static_cast<XProto::Gravity>(tmp16);

  // backing_planes
  Read(&backing_planes, &buf);

  // backing_pixel
  Read(&backing_pixel, &buf);

  // save_under
  Read(&save_under, &buf);

  // map_is_installed
  Read(&map_is_installed, &buf);

  // map_state
  uint8_t tmp17;
  Read(&tmp17, &buf);
  map_state = static_cast<XProto::MapState>(tmp17);

  // override_redirect
  Read(&override_redirect, &buf);

  // colormap
  Read(&colormap, &buf);

  // all_event_masks
  uint32_t tmp18;
  Read(&tmp18, &buf);
  all_event_masks = static_cast<XProto::EventMask>(tmp18);

  // your_event_mask
  uint32_t tmp19;
  Read(&tmp19, &buf);
  your_event_mask = static_cast<XProto::EventMask>(tmp19);

  // do_not_propagate_mask
  uint16_t tmp20;
  Read(&tmp20, &buf);
  do_not_propagate_mask = static_cast<XProto::EventMask>(tmp20);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::DestroyWindow(
    const XProto::DestroyWindowRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 4;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::DestroySubwindows(
    const XProto::DestroySubwindowsRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 5;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ChangeSaveSet(
    const XProto::ChangeSaveSetRequest& request) {
  WriteBuffer buf;

  auto& mode = request.mode;
  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 6;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp21;
  tmp21 = static_cast<uint8_t>(mode);
  Write(&tmp21, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ReparentWindow(
    const XProto::ReparentWindowRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;
  auto& parent = request.parent;
  auto& x = request.x;
  auto& y = request.y;

  // major_opcode
  uint8_t major_opcode = 7;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // parent
  Write(&parent, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::MapWindow(
    const XProto::MapWindowRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 8;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::MapSubwindows(
    const XProto::MapSubwindowsRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 9;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::UnmapWindow(
    const XProto::UnmapWindowRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 10;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::UnmapSubwindows(
    const XProto::UnmapSubwindowsRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 11;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ConfigureWindow(
    const XProto::ConfigureWindowRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;
  auto& value_mask = request.value_mask;
  auto& value_list = request.value_list;

  // major_opcode
  uint8_t major_opcode = 12;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // value_mask
  uint16_t tmp22;
  tmp22 = static_cast<uint16_t>(value_mask);
  Write(&tmp22, &buf);

  // pad1
  Pad(&buf, 2);

  // value_list
  {
    auto& x = value_list.x;
    auto& y = value_list.y;
    auto& width = value_list.width;
    auto& height = value_list.height;
    auto& border_width = value_list.border_width;
    auto& sibling = value_list.sibling;
    auto& stack_mode = value_list.stack_mode;

    auto value_list_expr = value_mask;
    if (BitAnd(value_list_expr, ConfigWindow::X)) {
      // x
      Write(&x, &buf);
    }
    if (BitAnd(value_list_expr, ConfigWindow::Y)) {
      // y
      Write(&y, &buf);
    }
    if (BitAnd(value_list_expr, ConfigWindow::Width)) {
      // width
      Write(&width, &buf);
    }
    if (BitAnd(value_list_expr, ConfigWindow::Height)) {
      // height
      Write(&height, &buf);
    }
    if (BitAnd(value_list_expr, ConfigWindow::BorderWidth)) {
      // border_width
      Write(&border_width, &buf);
    }
    if (BitAnd(value_list_expr, ConfigWindow::Sibling)) {
      // sibling
      Write(&sibling, &buf);
    }
    if (BitAnd(value_list_expr, ConfigWindow::StackMode)) {
      // stack_mode
      uint32_t tmp23;
      tmp23 = static_cast<uint32_t>(stack_mode);
      Write(&tmp23, &buf);
    }
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CirculateWindow(
    const XProto::CirculateWindowRequest& request) {
  WriteBuffer buf;

  auto& direction = request.direction;
  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 13;
  Write(&major_opcode, &buf);

  // direction
  uint8_t tmp24;
  tmp24 = static_cast<uint8_t>(direction);
  Write(&tmp24, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetGeometryReply>
XProto::GetGeometry(
    const XProto::GetGeometryRequest& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;

  // major_opcode
  uint8_t major_opcode = 14;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetGeometryReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetGeometryReply>
detail::ReadReply<XProto::GetGeometryReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetGeometryReply>();

  auto& depth = (*reply).depth;
  auto& root = (*reply).root;
  auto& x = (*reply).x;
  auto& y = (*reply).y;
  auto& width = (*reply).width;
  auto& height = (*reply).height;
  auto& border_width = (*reply).border_width;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // depth
  Read(&depth, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // root
  Read(&root, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // border_width
  Read(&border_width, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::QueryTreeReply>
XProto::QueryTree(
    const XProto::QueryTreeRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 15;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryTreeReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryTreeReply>
detail::ReadReply<XProto::QueryTreeReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryTreeReply>();

  auto& root = (*reply).root;
  auto& parent = (*reply).parent;
  uint16_t children_len;
  auto& children = (*reply).children;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // root
  Read(&root, &buf);

  // parent
  Read(&parent, &buf);

  // children_len
  Read(&children_len, &buf);

  // pad1
  Pad(&buf, 14);

  // children
  children.resize(children_len);
  for (auto& children_elem : children) {
    Read(&children_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::InternAtomReply>
XProto::InternAtom(
    const XProto::InternAtomRequest& request) {
  WriteBuffer buf;

  auto& only_if_exists = request.only_if_exists;
  uint16_t name_len;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 16;
  Write(&major_opcode, &buf);

  // only_if_exists
  Write(&only_if_exists, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad0
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    Write(&name_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::InternAtomReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::InternAtomReply>
detail::ReadReply<XProto::InternAtomReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::InternAtomReply>();

  auto& atom = (*reply).atom;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // atom
  Read(&atom, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::GetAtomNameReply>
XProto::GetAtomName(
    const XProto::GetAtomNameRequest& request) {
  WriteBuffer buf;

  auto& atom = request.atom;

  // major_opcode
  uint8_t major_opcode = 17;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // atom
  Write(&atom, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetAtomNameReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetAtomNameReply>
detail::ReadReply<XProto::GetAtomNameReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetAtomNameReply>();

  uint16_t name_len;
  auto& name = (*reply).name;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // name_len
  Read(&name_len, &buf);

  // pad1
  Pad(&buf, 22);

  // name
  name.resize(name_len);
  for (auto& name_elem : name) {
    Read(&name_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeProperty(
    const XProto::ChangePropertyRequest& request) {
  WriteBuffer buf;

  auto& mode = request.mode;
  auto& window = request.window;
  auto& property = request.property;
  auto& type = request.type;
  auto& format = request.format;
  auto& data_len = request.data_len;
  auto& data = request.data;

  // major_opcode
  uint8_t major_opcode = 18;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp25;
  tmp25 = static_cast<uint8_t>(mode);
  Write(&tmp25, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // property
  Write(&property, &buf);

  // type
  Write(&type, &buf);

  // format
  Write(&format, &buf);

  // pad0
  Pad(&buf, 3);

  // data_len
  Write(&data_len, &buf);

  // data
  DCHECK_EQ(static_cast<size_t>(((data_len) * (format)) / (8)), data.size());
  for (auto& data_elem : data) {
    Write(&data_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::DeleteProperty(
    const XProto::DeletePropertyRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;
  auto& property = request.property;

  // major_opcode
  uint8_t major_opcode = 19;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // property
  Write(&property, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetPropertyReply>
XProto::GetProperty(
    const XProto::GetPropertyRequest& request) {
  WriteBuffer buf;

  auto& c_delete = request.c_delete;
  auto& window = request.window;
  auto& property = request.property;
  auto& type = request.type;
  auto& long_offset = request.long_offset;
  auto& long_length = request.long_length;

  // major_opcode
  uint8_t major_opcode = 20;
  Write(&major_opcode, &buf);

  // c_delete
  Write(&c_delete, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // property
  Write(&property, &buf);

  // type
  Write(&type, &buf);

  // long_offset
  Write(&long_offset, &buf);

  // long_length
  Write(&long_length, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetPropertyReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetPropertyReply>
detail::ReadReply<XProto::GetPropertyReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetPropertyReply>();

  auto& format = (*reply).format;
  auto& type = (*reply).type;
  auto& bytes_after = (*reply).bytes_after;
  auto& value_len = (*reply).value_len;
  auto& value = (*reply).value;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // format
  Read(&format, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // type
  Read(&type, &buf);

  // bytes_after
  Read(&bytes_after, &buf);

  // value_len
  Read(&value_len, &buf);

  // pad0
  Pad(&buf, 12);

  // value
  value.resize((value_len) * ((format) / (8)));
  for (auto& value_elem : value) {
    Read(&value_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::ListPropertiesReply>
XProto::ListProperties(
    const XProto::ListPropertiesRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 21;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::ListPropertiesReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::ListPropertiesReply>
detail::ReadReply<XProto::ListPropertiesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::ListPropertiesReply>();

  auto& atoms_len = (*reply).atoms_len;
  auto& atoms = (*reply).atoms;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // atoms_len
  Read(&atoms_len, &buf);

  // pad1
  Pad(&buf, 22);

  // atoms
  atoms.resize(atoms_len);
  for (auto& atoms_elem : atoms) {
    Read(&atoms_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetSelectionOwner(
    const XProto::SetSelectionOwnerRequest& request) {
  WriteBuffer buf;

  auto& owner = request.owner;
  auto& selection = request.selection;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 22;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // owner
  Write(&owner, &buf);

  // selection
  Write(&selection, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetSelectionOwnerReply>
XProto::GetSelectionOwner(
    const XProto::GetSelectionOwnerRequest& request) {
  WriteBuffer buf;

  auto& selection = request.selection;

  // major_opcode
  uint8_t major_opcode = 23;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // selection
  Write(&selection, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetSelectionOwnerReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetSelectionOwnerReply>
detail::ReadReply<XProto::GetSelectionOwnerReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetSelectionOwnerReply>();

  auto& owner = (*reply).owner;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // owner
  Read(&owner, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ConvertSelection(
    const XProto::ConvertSelectionRequest& request) {
  WriteBuffer buf;

  auto& requestor = request.requestor;
  auto& selection = request.selection;
  auto& target = request.target;
  auto& property = request.property;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 24;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // requestor
  Write(&requestor, &buf);

  // selection
  Write(&selection, &buf);

  // target
  Write(&target, &buf);

  // property
  Write(&property, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::SendEvent(
    const XProto::SendEventRequest& request) {
  WriteBuffer buf;

  auto& propagate = request.propagate;
  auto& destination = request.destination;
  auto& event_mask = request.event_mask;
  auto& event = request.event;
  size_t event_len = event.size();

  // major_opcode
  uint8_t major_opcode = 25;
  Write(&major_opcode, &buf);

  // propagate
  Write(&propagate, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // destination
  Write(&destination, &buf);

  // event_mask
  uint32_t tmp26;
  tmp26 = static_cast<uint32_t>(event_mask);
  Write(&tmp26, &buf);

  // event
  for (auto& event_elem : event) {
    Write(&event_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GrabPointerReply>
XProto::GrabPointer(
    const XProto::GrabPointerRequest& request) {
  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& event_mask = request.event_mask;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;
  auto& confine_to = request.confine_to;
  auto& cursor = request.cursor;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 26;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // event_mask
  uint16_t tmp27;
  tmp27 = static_cast<uint16_t>(event_mask);
  Write(&tmp27, &buf);

  // pointer_mode
  uint8_t tmp28;
  tmp28 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp28, &buf);

  // keyboard_mode
  uint8_t tmp29;
  tmp29 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp29, &buf);

  // confine_to
  Write(&confine_to, &buf);

  // cursor
  Write(&cursor, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GrabPointerReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GrabPointerReply>
detail::ReadReply<XProto::GrabPointerReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GrabPointerReply>();

  auto& status = (*reply).status;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp30;
  Read(&tmp30, &buf);
  status = static_cast<XProto::GrabStatus>(tmp30);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::UngrabPointer(
    const XProto::UngrabPointerRequest& request) {
  WriteBuffer buf;

  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 27;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::GrabButton(
    const XProto::GrabButtonRequest& request) {
  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& event_mask = request.event_mask;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;
  auto& confine_to = request.confine_to;
  auto& cursor = request.cursor;
  auto& button = request.button;
  auto& modifiers = request.modifiers;

  // major_opcode
  uint8_t major_opcode = 28;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // event_mask
  uint16_t tmp31;
  tmp31 = static_cast<uint16_t>(event_mask);
  Write(&tmp31, &buf);

  // pointer_mode
  uint8_t tmp32;
  tmp32 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp32, &buf);

  // keyboard_mode
  uint8_t tmp33;
  tmp33 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp33, &buf);

  // confine_to
  Write(&confine_to, &buf);

  // cursor
  Write(&cursor, &buf);

  // button
  uint8_t tmp34;
  tmp34 = static_cast<uint8_t>(button);
  Write(&tmp34, &buf);

  // pad0
  Pad(&buf, 1);

  // modifiers
  uint16_t tmp35;
  tmp35 = static_cast<uint16_t>(modifiers);
  Write(&tmp35, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::UngrabButton(
    const XProto::UngrabButtonRequest& request) {
  WriteBuffer buf;

  auto& button = request.button;
  auto& grab_window = request.grab_window;
  auto& modifiers = request.modifiers;

  // major_opcode
  uint8_t major_opcode = 29;
  Write(&major_opcode, &buf);

  // button
  uint8_t tmp36;
  tmp36 = static_cast<uint8_t>(button);
  Write(&tmp36, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // modifiers
  uint16_t tmp37;
  tmp37 = static_cast<uint16_t>(modifiers);
  Write(&tmp37, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ChangeActivePointerGrab(
    const XProto::ChangeActivePointerGrabRequest& request) {
  WriteBuffer buf;

  auto& cursor = request.cursor;
  auto& time = request.time;
  auto& event_mask = request.event_mask;

  // major_opcode
  uint8_t major_opcode = 30;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cursor
  Write(&cursor, &buf);

  // time
  Write(&time, &buf);

  // event_mask
  uint16_t tmp38;
  tmp38 = static_cast<uint16_t>(event_mask);
  Write(&tmp38, &buf);

  // pad1
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GrabKeyboardReply>
XProto::GrabKeyboard(
    const XProto::GrabKeyboardRequest& request) {
  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& time = request.time;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;

  // major_opcode
  uint8_t major_opcode = 31;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // time
  Write(&time, &buf);

  // pointer_mode
  uint8_t tmp39;
  tmp39 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp39, &buf);

  // keyboard_mode
  uint8_t tmp40;
  tmp40 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp40, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GrabKeyboardReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GrabKeyboardReply>
detail::ReadReply<XProto::GrabKeyboardReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GrabKeyboardReply>();

  auto& status = (*reply).status;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp41;
  Read(&tmp41, &buf);
  status = static_cast<XProto::GrabStatus>(tmp41);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::UngrabKeyboard(
    const XProto::UngrabKeyboardRequest& request) {
  WriteBuffer buf;

  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 32;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::GrabKey(
    const XProto::GrabKeyRequest& request) {
  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& modifiers = request.modifiers;
  auto& key = request.key;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;

  // major_opcode
  uint8_t major_opcode = 33;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // modifiers
  uint16_t tmp42;
  tmp42 = static_cast<uint16_t>(modifiers);
  Write(&tmp42, &buf);

  // key
  Write(&key, &buf);

  // pointer_mode
  uint8_t tmp43;
  tmp43 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp43, &buf);

  // keyboard_mode
  uint8_t tmp44;
  tmp44 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp44, &buf);

  // pad0
  Pad(&buf, 3);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::UngrabKey(
    const XProto::UngrabKeyRequest& request) {
  WriteBuffer buf;

  auto& key = request.key;
  auto& grab_window = request.grab_window;
  auto& modifiers = request.modifiers;

  // major_opcode
  uint8_t major_opcode = 34;
  Write(&major_opcode, &buf);

  // key
  Write(&key, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // modifiers
  uint16_t tmp45;
  tmp45 = static_cast<uint16_t>(modifiers);
  Write(&tmp45, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::AllowEvents(
    const XProto::AllowEventsRequest& request) {
  WriteBuffer buf;

  auto& mode = request.mode;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 35;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp46;
  tmp46 = static_cast<uint8_t>(mode);
  Write(&tmp46, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::GrabServer(
    const XProto::GrabServerRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 36;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::UngrabServer(
    const XProto::UngrabServerRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 37;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::QueryPointerReply>
XProto::QueryPointer(
    const XProto::QueryPointerRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 38;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryPointerReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryPointerReply>
detail::ReadReply<XProto::QueryPointerReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryPointerReply>();

  auto& same_screen = (*reply).same_screen;
  auto& root = (*reply).root;
  auto& child = (*reply).child;
  auto& root_x = (*reply).root_x;
  auto& root_y = (*reply).root_y;
  auto& win_x = (*reply).win_x;
  auto& win_y = (*reply).win_y;
  auto& mask = (*reply).mask;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // same_screen
  Read(&same_screen, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // root
  Read(&root, &buf);

  // child
  Read(&child, &buf);

  // root_x
  Read(&root_x, &buf);

  // root_y
  Read(&root_y, &buf);

  // win_x
  Read(&win_x, &buf);

  // win_y
  Read(&win_y, &buf);

  // mask
  uint16_t tmp47;
  Read(&tmp47, &buf);
  mask = static_cast<XProto::KeyButMask>(tmp47);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::GetMotionEventsReply>
XProto::GetMotionEvents(
    const XProto::GetMotionEventsRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;
  auto& start = request.start;
  auto& stop = request.stop;

  // major_opcode
  uint8_t major_opcode = 39;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // start
  Write(&start, &buf);

  // stop
  Write(&stop, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetMotionEventsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetMotionEventsReply>
detail::ReadReply<XProto::GetMotionEventsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetMotionEventsReply>();

  uint32_t events_len;
  auto& events = (*reply).events;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // events_len
  Read(&events_len, &buf);

  // pad1
  Pad(&buf, 20);

  // events
  events.resize(events_len);
  for (auto& events_elem : events) {
    auto& time = events_elem.time;
    auto& x = events_elem.x;
    auto& y = events_elem.y;

    // time
    Read(&time, &buf);

    // x
    Read(&x, &buf);

    // y
    Read(&y, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::TranslateCoordinatesReply>
XProto::TranslateCoordinates(
    const XProto::TranslateCoordinatesRequest& request) {
  WriteBuffer buf;

  auto& src_window = request.src_window;
  auto& dst_window = request.dst_window;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;

  // major_opcode
  uint8_t major_opcode = 40;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_window
  Write(&src_window, &buf);

  // dst_window
  Write(&dst_window, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::TranslateCoordinatesReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::TranslateCoordinatesReply>
detail::ReadReply<XProto::TranslateCoordinatesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::TranslateCoordinatesReply>();

  auto& same_screen = (*reply).same_screen;
  auto& child = (*reply).child;
  auto& dst_x = (*reply).dst_x;
  auto& dst_y = (*reply).dst_y;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // same_screen
  Read(&same_screen, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // child
  Read(&child, &buf);

  // dst_x
  Read(&dst_x, &buf);

  // dst_y
  Read(&dst_y, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::WarpPointer(
    const XProto::WarpPointerRequest& request) {
  WriteBuffer buf;

  auto& src_window = request.src_window;
  auto& dst_window = request.dst_window;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& src_width = request.src_width;
  auto& src_height = request.src_height;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;

  // major_opcode
  uint8_t major_opcode = 41;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_window
  Write(&src_window, &buf);

  // dst_window
  Write(&dst_window, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // src_width
  Write(&src_width, &buf);

  // src_height
  Write(&src_height, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::SetInputFocus(
    const XProto::SetInputFocusRequest& request) {
  WriteBuffer buf;

  auto& revert_to = request.revert_to;
  auto& focus = request.focus;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 42;
  Write(&major_opcode, &buf);

  // revert_to
  uint8_t tmp48;
  tmp48 = static_cast<uint8_t>(revert_to);
  Write(&tmp48, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // focus
  Write(&focus, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetInputFocusReply>
XProto::GetInputFocus(
    const XProto::GetInputFocusRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 43;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetInputFocusReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetInputFocusReply>
detail::ReadReply<XProto::GetInputFocusReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetInputFocusReply>();

  auto& revert_to = (*reply).revert_to;
  auto& focus = (*reply).focus;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // revert_to
  uint8_t tmp49;
  Read(&tmp49, &buf);
  revert_to = static_cast<XProto::InputFocus>(tmp49);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // focus
  Read(&focus, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::QueryKeymapReply>
XProto::QueryKeymap(
    const XProto::QueryKeymapRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 44;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryKeymapReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryKeymapReply>
detail::ReadReply<XProto::QueryKeymapReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryKeymapReply>();

  auto& keys = (*reply).keys;
  size_t keys_len = keys.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // keys
  for (auto& keys_elem : keys) {
    Read(&keys_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::OpenFont(
    const XProto::OpenFontRequest& request) {
  WriteBuffer buf;

  auto& fid = request.fid;
  uint16_t name_len;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 45;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // fid
  Write(&fid, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    Write(&name_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CloseFont(
    const XProto::CloseFontRequest& request) {
  WriteBuffer buf;

  auto& font = request.font;

  // major_opcode
  uint8_t major_opcode = 46;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font
  Write(&font, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::QueryFontReply>
XProto::QueryFont(
    const XProto::QueryFontRequest& request) {
  WriteBuffer buf;

  auto& font = request.font;

  // major_opcode
  uint8_t major_opcode = 47;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font
  Write(&font, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryFontReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryFontReply>
detail::ReadReply<XProto::QueryFontReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryFontReply>();

  auto& min_bounds = (*reply).min_bounds;
  auto& max_bounds = (*reply).max_bounds;
  auto& min_char_or_byte2 = (*reply).min_char_or_byte2;
  auto& max_char_or_byte2 = (*reply).max_char_or_byte2;
  auto& default_char = (*reply).default_char;
  uint16_t properties_len;
  auto& draw_direction = (*reply).draw_direction;
  auto& min_byte1 = (*reply).min_byte1;
  auto& max_byte1 = (*reply).max_byte1;
  auto& all_chars_exist = (*reply).all_chars_exist;
  auto& font_ascent = (*reply).font_ascent;
  auto& font_descent = (*reply).font_descent;
  uint32_t char_infos_len;
  auto& properties = (*reply).properties;
  auto& char_infos = (*reply).char_infos;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // min_bounds
  {
    auto& left_side_bearing = min_bounds.left_side_bearing;
    auto& right_side_bearing = min_bounds.right_side_bearing;
    auto& character_width = min_bounds.character_width;
    auto& ascent = min_bounds.ascent;
    auto& descent = min_bounds.descent;
    auto& attributes = min_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad1
  Pad(&buf, 4);

  // max_bounds
  {
    auto& left_side_bearing = max_bounds.left_side_bearing;
    auto& right_side_bearing = max_bounds.right_side_bearing;
    auto& character_width = max_bounds.character_width;
    auto& ascent = max_bounds.ascent;
    auto& descent = max_bounds.descent;
    auto& attributes = max_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad2
  Pad(&buf, 4);

  // min_char_or_byte2
  Read(&min_char_or_byte2, &buf);

  // max_char_or_byte2
  Read(&max_char_or_byte2, &buf);

  // default_char
  Read(&default_char, &buf);

  // properties_len
  Read(&properties_len, &buf);

  // draw_direction
  uint8_t tmp50;
  Read(&tmp50, &buf);
  draw_direction = static_cast<XProto::FontDraw>(tmp50);

  // min_byte1
  Read(&min_byte1, &buf);

  // max_byte1
  Read(&max_byte1, &buf);

  // all_chars_exist
  Read(&all_chars_exist, &buf);

  // font_ascent
  Read(&font_ascent, &buf);

  // font_descent
  Read(&font_descent, &buf);

  // char_infos_len
  Read(&char_infos_len, &buf);

  // properties
  properties.resize(properties_len);
  for (auto& properties_elem : properties) {
    auto& name = properties_elem.name;
    auto& value = properties_elem.value;

    // name
    Read(&name, &buf);

    // value
    Read(&value, &buf);

  }

  // char_infos
  char_infos.resize(char_infos_len);
  for (auto& char_infos_elem : char_infos) {
    auto& left_side_bearing = char_infos_elem.left_side_bearing;
    auto& right_side_bearing = char_infos_elem.right_side_bearing;
    auto& character_width = char_infos_elem.character_width;
    auto& ascent = char_infos_elem.ascent;
    auto& descent = char_infos_elem.descent;
    auto& attributes = char_infos_elem.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::QueryTextExtentsReply>
XProto::QueryTextExtents(
    const XProto::QueryTextExtentsRequest& request) {
  WriteBuffer buf;

  auto& font = request.font;
  auto& string = request.string;
  size_t string_len = string.size();

  // major_opcode
  uint8_t major_opcode = 48;
  Write(&major_opcode, &buf);

  // odd_length
  uint8_t odd_length = BitAnd(string_len, 1);
  Write(&odd_length, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font
  Write(&font, &buf);

  // string
  DCHECK_EQ(static_cast<size_t>(string_len), string.size());
  for (auto& string_elem : string) {
    auto& byte1 = string_elem.byte1;
    auto& byte2 = string_elem.byte2;

    // byte1
    Write(&byte1, &buf);

    // byte2
    Write(&byte2, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryTextExtentsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryTextExtentsReply>
detail::ReadReply<XProto::QueryTextExtentsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryTextExtentsReply>();

  auto& draw_direction = (*reply).draw_direction;
  auto& font_ascent = (*reply).font_ascent;
  auto& font_descent = (*reply).font_descent;
  auto& overall_ascent = (*reply).overall_ascent;
  auto& overall_descent = (*reply).overall_descent;
  auto& overall_width = (*reply).overall_width;
  auto& overall_left = (*reply).overall_left;
  auto& overall_right = (*reply).overall_right;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // draw_direction
  uint8_t tmp51;
  Read(&tmp51, &buf);
  draw_direction = static_cast<XProto::FontDraw>(tmp51);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // font_ascent
  Read(&font_ascent, &buf);

  // font_descent
  Read(&font_descent, &buf);

  // overall_ascent
  Read(&overall_ascent, &buf);

  // overall_descent
  Read(&overall_descent, &buf);

  // overall_width
  Read(&overall_width, &buf);

  // overall_left
  Read(&overall_left, &buf);

  // overall_right
  Read(&overall_right, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::ListFontsReply>
XProto::ListFonts(
    const XProto::ListFontsRequest& request) {
  WriteBuffer buf;

  auto& max_names = request.max_names;
  uint16_t pattern_len;
  auto& pattern = request.pattern;

  // major_opcode
  uint8_t major_opcode = 49;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // max_names
  Write(&max_names, &buf);

  // pattern_len
  pattern_len = pattern.size();
  Write(&pattern_len, &buf);

  // pattern
  DCHECK_EQ(static_cast<size_t>(pattern_len), pattern.size());
  for (auto& pattern_elem : pattern) {
    Write(&pattern_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::ListFontsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::ListFontsReply>
detail::ReadReply<XProto::ListFontsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::ListFontsReply>();

  auto& names_len = (*reply).names_len;
  auto& names = (*reply).names;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // names_len
  Read(&names_len, &buf);

  // pad1
  Pad(&buf, 22);

  // names
  names.resize(names_len);
  for (auto& names_elem : names) {
    uint8_t name_len;
    auto& name = names_elem.name;

    // name_len
    Read(&name_len, &buf);

    // name
    name.resize(name_len);
    for (auto& name_elem : name) {
      Read(&name_elem, &buf);
    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::ListFontsWithInfoReply>
XProto::ListFontsWithInfo(
    const XProto::ListFontsWithInfoRequest& request) {
  WriteBuffer buf;

  auto& max_names = request.max_names;
  uint16_t pattern_len;
  auto& pattern = request.pattern;

  // major_opcode
  uint8_t major_opcode = 50;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // max_names
  Write(&max_names, &buf);

  // pattern_len
  pattern_len = pattern.size();
  Write(&pattern_len, &buf);

  // pattern
  DCHECK_EQ(static_cast<size_t>(pattern_len), pattern.size());
  for (auto& pattern_elem : pattern) {
    Write(&pattern_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::ListFontsWithInfoReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::ListFontsWithInfoReply>
detail::ReadReply<XProto::ListFontsWithInfoReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::ListFontsWithInfoReply>();

  uint8_t name_len;
  auto& min_bounds = (*reply).min_bounds;
  auto& max_bounds = (*reply).max_bounds;
  auto& min_char_or_byte2 = (*reply).min_char_or_byte2;
  auto& max_char_or_byte2 = (*reply).max_char_or_byte2;
  auto& default_char = (*reply).default_char;
  uint16_t properties_len;
  auto& draw_direction = (*reply).draw_direction;
  auto& min_byte1 = (*reply).min_byte1;
  auto& max_byte1 = (*reply).max_byte1;
  auto& all_chars_exist = (*reply).all_chars_exist;
  auto& font_ascent = (*reply).font_ascent;
  auto& font_descent = (*reply).font_descent;
  auto& replies_hint = (*reply).replies_hint;
  auto& properties = (*reply).properties;
  auto& name = (*reply).name;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // name_len
  Read(&name_len, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // min_bounds
  {
    auto& left_side_bearing = min_bounds.left_side_bearing;
    auto& right_side_bearing = min_bounds.right_side_bearing;
    auto& character_width = min_bounds.character_width;
    auto& ascent = min_bounds.ascent;
    auto& descent = min_bounds.descent;
    auto& attributes = min_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad0
  Pad(&buf, 4);

  // max_bounds
  {
    auto& left_side_bearing = max_bounds.left_side_bearing;
    auto& right_side_bearing = max_bounds.right_side_bearing;
    auto& character_width = max_bounds.character_width;
    auto& ascent = max_bounds.ascent;
    auto& descent = max_bounds.descent;
    auto& attributes = max_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad1
  Pad(&buf, 4);

  // min_char_or_byte2
  Read(&min_char_or_byte2, &buf);

  // max_char_or_byte2
  Read(&max_char_or_byte2, &buf);

  // default_char
  Read(&default_char, &buf);

  // properties_len
  Read(&properties_len, &buf);

  // draw_direction
  uint8_t tmp52;
  Read(&tmp52, &buf);
  draw_direction = static_cast<XProto::FontDraw>(tmp52);

  // min_byte1
  Read(&min_byte1, &buf);

  // max_byte1
  Read(&max_byte1, &buf);

  // all_chars_exist
  Read(&all_chars_exist, &buf);

  // font_ascent
  Read(&font_ascent, &buf);

  // font_descent
  Read(&font_descent, &buf);

  // replies_hint
  Read(&replies_hint, &buf);

  // properties
  properties.resize(properties_len);
  for (auto& properties_elem : properties) {
    auto& name = properties_elem.name;
    auto& value = properties_elem.value;

    // name
    Read(&name, &buf);

    // value
    Read(&value, &buf);

  }

  // name
  name.resize(name_len);
  for (auto& name_elem : name) {
    Read(&name_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetFontPath(
    const XProto::SetFontPathRequest& request) {
  WriteBuffer buf;

  uint16_t font_qty;
  auto& font = request.font;
  size_t font_len = font.size();

  // major_opcode
  uint8_t major_opcode = 51;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font_qty
  font_qty = font.size();
  Write(&font_qty, &buf);

  // pad1
  Pad(&buf, 2);

  // font
  DCHECK_EQ(static_cast<size_t>(font_qty), font.size());
  for (auto& font_elem : font) {
    uint8_t name_len;
    auto& name = font_elem.name;

    // name_len
    name_len = name.size();
    Write(&name_len, &buf);

    // name
    DCHECK_EQ(static_cast<size_t>(name_len), name.size());
    for (auto& name_elem : name) {
      Write(&name_elem, &buf);
    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetFontPathReply>
XProto::GetFontPath(
    const XProto::GetFontPathRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 52;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetFontPathReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetFontPathReply>
detail::ReadReply<XProto::GetFontPathReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetFontPathReply>();

  uint16_t path_len;
  auto& path = (*reply).path;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // path_len
  Read(&path_len, &buf);

  // pad1
  Pad(&buf, 22);

  // path
  path.resize(path_len);
  for (auto& path_elem : path) {
    uint8_t name_len;
    auto& name = path_elem.name;

    // name_len
    Read(&name_len, &buf);

    // name
    name.resize(name_len);
    for (auto& name_elem : name) {
      Read(&name_elem, &buf);
    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::CreatePixmap(
    const XProto::CreatePixmapRequest& request) {
  WriteBuffer buf;

  auto& depth = request.depth;
  auto& pid = request.pid;
  auto& drawable = request.drawable;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 53;
  Write(&major_opcode, &buf);

  // depth
  Write(&depth, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pid
  Write(&pid, &buf);

  // drawable
  Write(&drawable, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::FreePixmap(
    const XProto::FreePixmapRequest& request) {
  WriteBuffer buf;

  auto& pixmap = request.pixmap;

  // major_opcode
  uint8_t major_opcode = 54;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pixmap
  Write(&pixmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CreateGC(
    const XProto::CreateGCRequest& request) {
  WriteBuffer buf;

  auto& cid = request.cid;
  auto& drawable = request.drawable;
  auto& value_mask = request.value_mask;
  auto& value_list = request.value_list;

  // major_opcode
  uint8_t major_opcode = 55;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // drawable
  Write(&drawable, &buf);

  // value_mask
  uint32_t tmp53;
  tmp53 = static_cast<uint32_t>(value_mask);
  Write(&tmp53, &buf);

  // value_list
  {
    auto& function = value_list.function;
    auto& plane_mask = value_list.plane_mask;
    auto& foreground = value_list.foreground;
    auto& background = value_list.background;
    auto& line_width = value_list.line_width;
    auto& line_style = value_list.line_style;
    auto& cap_style = value_list.cap_style;
    auto& join_style = value_list.join_style;
    auto& fill_style = value_list.fill_style;
    auto& fill_rule = value_list.fill_rule;
    auto& tile = value_list.tile;
    auto& stipple = value_list.stipple;
    auto& tile_stipple_x_origin = value_list.tile_stipple_x_origin;
    auto& tile_stipple_y_origin = value_list.tile_stipple_y_origin;
    auto& font = value_list.font;
    auto& subwindow_mode = value_list.subwindow_mode;
    auto& graphics_exposures = value_list.graphics_exposures;
    auto& clip_x_origin = value_list.clip_x_origin;
    auto& clip_y_origin = value_list.clip_y_origin;
    auto& clip_mask = value_list.clip_mask;
    auto& dash_offset = value_list.dash_offset;
    auto& dashes = value_list.dashes;
    auto& arc_mode = value_list.arc_mode;

    auto value_list_expr = value_mask;
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Function)) {
      // function
      uint32_t tmp54;
      tmp54 = static_cast<uint32_t>(function);
      Write(&tmp54, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::PlaneMask)) {
      // plane_mask
      Write(&plane_mask, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Foreground)) {
      // foreground
      Write(&foreground, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Background)) {
      // background
      Write(&background, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::LineWidth)) {
      // line_width
      Write(&line_width, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::LineStyle)) {
      // line_style
      uint32_t tmp55;
      tmp55 = static_cast<uint32_t>(line_style);
      Write(&tmp55, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::CapStyle)) {
      // cap_style
      uint32_t tmp56;
      tmp56 = static_cast<uint32_t>(cap_style);
      Write(&tmp56, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::JoinStyle)) {
      // join_style
      uint32_t tmp57;
      tmp57 = static_cast<uint32_t>(join_style);
      Write(&tmp57, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::FillStyle)) {
      // fill_style
      uint32_t tmp58;
      tmp58 = static_cast<uint32_t>(fill_style);
      Write(&tmp58, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::FillRule)) {
      // fill_rule
      uint32_t tmp59;
      tmp59 = static_cast<uint32_t>(fill_rule);
      Write(&tmp59, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Tile)) {
      // tile
      Write(&tile, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Stipple)) {
      // stipple
      Write(&stipple, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginX)) {
      // tile_stipple_x_origin
      Write(&tile_stipple_x_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginY)) {
      // tile_stipple_y_origin
      Write(&tile_stipple_y_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Font)) {
      // font
      Write(&font, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::SubwindowMode)) {
      // subwindow_mode
      uint32_t tmp60;
      tmp60 = static_cast<uint32_t>(subwindow_mode);
      Write(&tmp60, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::GraphicsExposures)) {
      // graphics_exposures
      Write(&graphics_exposures, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ClipOriginX)) {
      // clip_x_origin
      Write(&clip_x_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ClipOriginY)) {
      // clip_y_origin
      Write(&clip_y_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ClipMask)) {
      // clip_mask
      Write(&clip_mask, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::DashOffset)) {
      // dash_offset
      Write(&dash_offset, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::DashList)) {
      // dashes
      Write(&dashes, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ArcMode)) {
      // arc_mode
      uint32_t tmp61;
      tmp61 = static_cast<uint32_t>(arc_mode);
      Write(&tmp61, &buf);
    }
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ChangeGC(
    const XProto::ChangeGCRequest& request) {
  WriteBuffer buf;

  auto& gc = request.gc;
  auto& value_mask = request.value_mask;
  auto& value_list = request.value_list;

  // major_opcode
  uint8_t major_opcode = 56;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  // value_mask
  uint32_t tmp62;
  tmp62 = static_cast<uint32_t>(value_mask);
  Write(&tmp62, &buf);

  // value_list
  {
    auto& function = value_list.function;
    auto& plane_mask = value_list.plane_mask;
    auto& foreground = value_list.foreground;
    auto& background = value_list.background;
    auto& line_width = value_list.line_width;
    auto& line_style = value_list.line_style;
    auto& cap_style = value_list.cap_style;
    auto& join_style = value_list.join_style;
    auto& fill_style = value_list.fill_style;
    auto& fill_rule = value_list.fill_rule;
    auto& tile = value_list.tile;
    auto& stipple = value_list.stipple;
    auto& tile_stipple_x_origin = value_list.tile_stipple_x_origin;
    auto& tile_stipple_y_origin = value_list.tile_stipple_y_origin;
    auto& font = value_list.font;
    auto& subwindow_mode = value_list.subwindow_mode;
    auto& graphics_exposures = value_list.graphics_exposures;
    auto& clip_x_origin = value_list.clip_x_origin;
    auto& clip_y_origin = value_list.clip_y_origin;
    auto& clip_mask = value_list.clip_mask;
    auto& dash_offset = value_list.dash_offset;
    auto& dashes = value_list.dashes;
    auto& arc_mode = value_list.arc_mode;

    auto value_list_expr = value_mask;
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Function)) {
      // function
      uint32_t tmp63;
      tmp63 = static_cast<uint32_t>(function);
      Write(&tmp63, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::PlaneMask)) {
      // plane_mask
      Write(&plane_mask, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Foreground)) {
      // foreground
      Write(&foreground, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Background)) {
      // background
      Write(&background, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::LineWidth)) {
      // line_width
      Write(&line_width, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::LineStyle)) {
      // line_style
      uint32_t tmp64;
      tmp64 = static_cast<uint32_t>(line_style);
      Write(&tmp64, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::CapStyle)) {
      // cap_style
      uint32_t tmp65;
      tmp65 = static_cast<uint32_t>(cap_style);
      Write(&tmp65, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::JoinStyle)) {
      // join_style
      uint32_t tmp66;
      tmp66 = static_cast<uint32_t>(join_style);
      Write(&tmp66, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::FillStyle)) {
      // fill_style
      uint32_t tmp67;
      tmp67 = static_cast<uint32_t>(fill_style);
      Write(&tmp67, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::FillRule)) {
      // fill_rule
      uint32_t tmp68;
      tmp68 = static_cast<uint32_t>(fill_rule);
      Write(&tmp68, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Tile)) {
      // tile
      Write(&tile, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Stipple)) {
      // stipple
      Write(&stipple, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginX)) {
      // tile_stipple_x_origin
      Write(&tile_stipple_x_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginY)) {
      // tile_stipple_y_origin
      Write(&tile_stipple_y_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::Font)) {
      // font
      Write(&font, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::SubwindowMode)) {
      // subwindow_mode
      uint32_t tmp69;
      tmp69 = static_cast<uint32_t>(subwindow_mode);
      Write(&tmp69, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::GraphicsExposures)) {
      // graphics_exposures
      Write(&graphics_exposures, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ClipOriginX)) {
      // clip_x_origin
      Write(&clip_x_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ClipOriginY)) {
      // clip_y_origin
      Write(&clip_y_origin, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ClipMask)) {
      // clip_mask
      Write(&clip_mask, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::DashOffset)) {
      // dash_offset
      Write(&dash_offset, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::DashList)) {
      // dashes
      Write(&dashes, &buf);
    }
    if (BitAnd(value_list_expr, GraphicsContextAttribute::ArcMode)) {
      // arc_mode
      uint32_t tmp70;
      tmp70 = static_cast<uint32_t>(arc_mode);
      Write(&tmp70, &buf);
    }
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CopyGC(
    const XProto::CopyGCRequest& request) {
  WriteBuffer buf;

  auto& src_gc = request.src_gc;
  auto& dst_gc = request.dst_gc;
  auto& value_mask = request.value_mask;

  // major_opcode
  uint8_t major_opcode = 57;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_gc
  Write(&src_gc, &buf);

  // dst_gc
  Write(&dst_gc, &buf);

  // value_mask
  uint32_t tmp71;
  tmp71 = static_cast<uint32_t>(value_mask);
  Write(&tmp71, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::SetDashes(
    const XProto::SetDashesRequest& request) {
  WriteBuffer buf;

  auto& gc = request.gc;
  auto& dash_offset = request.dash_offset;
  auto& dashes_len = request.dashes_len;
  auto& dashes = request.dashes;

  // major_opcode
  uint8_t major_opcode = 58;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  // dash_offset
  Write(&dash_offset, &buf);

  // dashes_len
  Write(&dashes_len, &buf);

  // dashes
  DCHECK_EQ(static_cast<size_t>(dashes_len), dashes.size());
  for (auto& dashes_elem : dashes) {
    Write(&dashes_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::SetClipRectangles(
    const XProto::SetClipRectanglesRequest& request) {
  WriteBuffer buf;

  auto& ordering = request.ordering;
  auto& gc = request.gc;
  auto& clip_x_origin = request.clip_x_origin;
  auto& clip_y_origin = request.clip_y_origin;
  auto& rectangles = request.rectangles;
  size_t rectangles_len = rectangles.size();

  // major_opcode
  uint8_t major_opcode = 59;
  Write(&major_opcode, &buf);

  // ordering
  uint8_t tmp72;
  tmp72 = static_cast<uint8_t>(ordering);
  Write(&tmp72, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  // clip_x_origin
  Write(&clip_x_origin, &buf);

  // clip_y_origin
  Write(&clip_y_origin, &buf);

  // rectangles
  DCHECK_EQ(static_cast<size_t>(rectangles_len), rectangles.size());
  for (auto& rectangles_elem : rectangles) {
    auto& x = rectangles_elem.x;
    auto& y = rectangles_elem.y;
    auto& width = rectangles_elem.width;
    auto& height = rectangles_elem.height;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

    // width
    Write(&width, &buf);

    // height
    Write(&height, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::FreeGC(
    const XProto::FreeGCRequest& request) {
  WriteBuffer buf;

  auto& gc = request.gc;

  // major_opcode
  uint8_t major_opcode = 60;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ClearArea(
    const XProto::ClearAreaRequest& request) {
  WriteBuffer buf;

  auto& exposures = request.exposures;
  auto& window = request.window;
  auto& x = request.x;
  auto& y = request.y;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 61;
  Write(&major_opcode, &buf);

  // exposures
  Write(&exposures, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CopyArea(
    const XProto::CopyAreaRequest& request) {
  WriteBuffer buf;

  auto& src_drawable = request.src_drawable;
  auto& dst_drawable = request.dst_drawable;
  auto& gc = request.gc;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 62;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_drawable
  Write(&src_drawable, &buf);

  // dst_drawable
  Write(&dst_drawable, &buf);

  // gc
  Write(&gc, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CopyPlane(
    const XProto::CopyPlaneRequest& request) {
  WriteBuffer buf;

  auto& src_drawable = request.src_drawable;
  auto& dst_drawable = request.dst_drawable;
  auto& gc = request.gc;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;
  auto& width = request.width;
  auto& height = request.height;
  auto& bit_plane = request.bit_plane;

  // major_opcode
  uint8_t major_opcode = 63;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_drawable
  Write(&src_drawable, &buf);

  // dst_drawable
  Write(&dst_drawable, &buf);

  // gc
  Write(&gc, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // bit_plane
  Write(&bit_plane, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolyPoint(
    const XProto::PolyPointRequest& request) {
  WriteBuffer buf;

  auto& coordinate_mode = request.coordinate_mode;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = 64;
  Write(&major_opcode, &buf);

  // coordinate_mode
  uint8_t tmp73;
  tmp73 = static_cast<uint8_t>(coordinate_mode);
  Write(&tmp73, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    auto& x = points_elem.x;
    auto& y = points_elem.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolyLine(
    const XProto::PolyLineRequest& request) {
  WriteBuffer buf;

  auto& coordinate_mode = request.coordinate_mode;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = 65;
  Write(&major_opcode, &buf);

  // coordinate_mode
  uint8_t tmp74;
  tmp74 = static_cast<uint8_t>(coordinate_mode);
  Write(&tmp74, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    auto& x = points_elem.x;
    auto& y = points_elem.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolySegment(
    const XProto::PolySegmentRequest& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& segments = request.segments;
  size_t segments_len = segments.size();

  // major_opcode
  uint8_t major_opcode = 66;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // segments
  DCHECK_EQ(static_cast<size_t>(segments_len), segments.size());
  for (auto& segments_elem : segments) {
    auto& x1 = segments_elem.x1;
    auto& y1 = segments_elem.y1;
    auto& x2 = segments_elem.x2;
    auto& y2 = segments_elem.y2;

    // x1
    Write(&x1, &buf);

    // y1
    Write(&y1, &buf);

    // x2
    Write(&x2, &buf);

    // y2
    Write(&y2, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolyRectangle(
    const XProto::PolyRectangleRequest& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& rectangles = request.rectangles;
  size_t rectangles_len = rectangles.size();

  // major_opcode
  uint8_t major_opcode = 67;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // rectangles
  DCHECK_EQ(static_cast<size_t>(rectangles_len), rectangles.size());
  for (auto& rectangles_elem : rectangles) {
    auto& x = rectangles_elem.x;
    auto& y = rectangles_elem.y;
    auto& width = rectangles_elem.width;
    auto& height = rectangles_elem.height;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

    // width
    Write(&width, &buf);

    // height
    Write(&height, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolyArc(
    const XProto::PolyArcRequest& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& arcs = request.arcs;
  size_t arcs_len = arcs.size();

  // major_opcode
  uint8_t major_opcode = 68;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // arcs
  DCHECK_EQ(static_cast<size_t>(arcs_len), arcs.size());
  for (auto& arcs_elem : arcs) {
    auto& x = arcs_elem.x;
    auto& y = arcs_elem.y;
    auto& width = arcs_elem.width;
    auto& height = arcs_elem.height;
    auto& angle1 = arcs_elem.angle1;
    auto& angle2 = arcs_elem.angle2;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

    // width
    Write(&width, &buf);

    // height
    Write(&height, &buf);

    // angle1
    Write(&angle1, &buf);

    // angle2
    Write(&angle2, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::FillPoly(
    const XProto::FillPolyRequest& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& shape = request.shape;
  auto& coordinate_mode = request.coordinate_mode;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = 69;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // shape
  uint8_t tmp75;
  tmp75 = static_cast<uint8_t>(shape);
  Write(&tmp75, &buf);

  // coordinate_mode
  uint8_t tmp76;
  tmp76 = static_cast<uint8_t>(coordinate_mode);
  Write(&tmp76, &buf);

  // pad1
  Pad(&buf, 2);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    auto& x = points_elem.x;
    auto& y = points_elem.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolyFillRectangle(
    const XProto::PolyFillRectangleRequest& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& rectangles = request.rectangles;
  size_t rectangles_len = rectangles.size();

  // major_opcode
  uint8_t major_opcode = 70;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // rectangles
  DCHECK_EQ(static_cast<size_t>(rectangles_len), rectangles.size());
  for (auto& rectangles_elem : rectangles) {
    auto& x = rectangles_elem.x;
    auto& y = rectangles_elem.y;
    auto& width = rectangles_elem.width;
    auto& height = rectangles_elem.height;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

    // width
    Write(&width, &buf);

    // height
    Write(&height, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolyFillArc(
    const XProto::PolyFillArcRequest& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& arcs = request.arcs;
  size_t arcs_len = arcs.size();

  // major_opcode
  uint8_t major_opcode = 71;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // arcs
  DCHECK_EQ(static_cast<size_t>(arcs_len), arcs.size());
  for (auto& arcs_elem : arcs) {
    auto& x = arcs_elem.x;
    auto& y = arcs_elem.y;
    auto& width = arcs_elem.width;
    auto& height = arcs_elem.height;
    auto& angle1 = arcs_elem.angle1;
    auto& angle2 = arcs_elem.angle2;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

    // width
    Write(&width, &buf);

    // height
    Write(&height, &buf);

    // angle1
    Write(&angle1, &buf);

    // angle2
    Write(&angle2, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PutImage(
    const XProto::PutImageRequest& request) {
  WriteBuffer buf;

  auto& format = request.format;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& width = request.width;
  auto& height = request.height;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;
  auto& left_pad = request.left_pad;
  auto& depth = request.depth;
  auto& data = request.data;
  size_t data_len = data.size();

  // major_opcode
  uint8_t major_opcode = 72;
  Write(&major_opcode, &buf);

  // format
  uint8_t tmp77;
  tmp77 = static_cast<uint8_t>(format);
  Write(&tmp77, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  // left_pad
  Write(&left_pad, &buf);

  // depth
  Write(&depth, &buf);

  // pad0
  Pad(&buf, 2);

  // data
  DCHECK_EQ(static_cast<size_t>(data_len), data.size());
  for (auto& data_elem : data) {
    Write(&data_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetImageReply>
XProto::GetImage(
    const XProto::GetImageRequest& request) {
  WriteBuffer buf;

  auto& format = request.format;
  auto& drawable = request.drawable;
  auto& x = request.x;
  auto& y = request.y;
  auto& width = request.width;
  auto& height = request.height;
  auto& plane_mask = request.plane_mask;

  // major_opcode
  uint8_t major_opcode = 73;
  Write(&major_opcode, &buf);

  // format
  uint8_t tmp78;
  tmp78 = static_cast<uint8_t>(format);
  Write(&tmp78, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // plane_mask
  Write(&plane_mask, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetImageReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetImageReply>
detail::ReadReply<XProto::GetImageReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetImageReply>();

  auto& depth = (*reply).depth;
  auto& visual = (*reply).visual;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // depth
  Read(&depth, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // visual
  Read(&visual, &buf);

  // pad0
  Pad(&buf, 20);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    Read(&data_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::PolyText8(
    const XProto::PolyText8Request& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& items = request.items;
  size_t items_len = items.size();

  // major_opcode
  uint8_t major_opcode = 74;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // items
  DCHECK_EQ(static_cast<size_t>(items_len), items.size());
  for (auto& items_elem : items) {
    Write(&items_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::PolyText16(
    const XProto::PolyText16Request& request) {
  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& items = request.items;
  size_t items_len = items.size();

  // major_opcode
  uint8_t major_opcode = 75;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // items
  DCHECK_EQ(static_cast<size_t>(items_len), items.size());
  for (auto& items_elem : items) {
    Write(&items_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ImageText8(
    const XProto::ImageText8Request& request) {
  WriteBuffer buf;

  auto& string_len = request.string_len;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& string = request.string;

  // major_opcode
  uint8_t major_opcode = 76;
  Write(&major_opcode, &buf);

  // string_len
  Write(&string_len, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // string
  DCHECK_EQ(static_cast<size_t>(string_len), string.size());
  for (auto& string_elem : string) {
    Write(&string_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ImageText16(
    const XProto::ImageText16Request& request) {
  WriteBuffer buf;

  auto& string_len = request.string_len;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& string = request.string;

  // major_opcode
  uint8_t major_opcode = 77;
  Write(&major_opcode, &buf);

  // string_len
  Write(&string_len, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // string
  DCHECK_EQ(static_cast<size_t>(string_len), string.size());
  for (auto& string_elem : string) {
    auto& byte1 = string_elem.byte1;
    auto& byte2 = string_elem.byte2;

    // byte1
    Write(&byte1, &buf);

    // byte2
    Write(&byte2, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CreateColormap(
    const XProto::CreateColormapRequest& request) {
  WriteBuffer buf;

  auto& alloc = request.alloc;
  auto& mid = request.mid;
  auto& window = request.window;
  auto& visual = request.visual;

  // major_opcode
  uint8_t major_opcode = 78;
  Write(&major_opcode, &buf);

  // alloc
  uint8_t tmp79;
  tmp79 = static_cast<uint8_t>(alloc);
  Write(&tmp79, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // mid
  Write(&mid, &buf);

  // window
  Write(&window, &buf);

  // visual
  Write(&visual, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::FreeColormap(
    const XProto::FreeColormapRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;

  // major_opcode
  uint8_t major_opcode = 79;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CopyColormapAndFree(
    const XProto::CopyColormapAndFreeRequest& request) {
  WriteBuffer buf;

  auto& mid = request.mid;
  auto& src_cmap = request.src_cmap;

  // major_opcode
  uint8_t major_opcode = 80;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // mid
  Write(&mid, &buf);

  // src_cmap
  Write(&src_cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::InstallColormap(
    const XProto::InstallColormapRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;

  // major_opcode
  uint8_t major_opcode = 81;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::UninstallColormap(
    const XProto::UninstallColormapRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;

  // major_opcode
  uint8_t major_opcode = 82;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::ListInstalledColormapsReply>
XProto::ListInstalledColormaps(
    const XProto::ListInstalledColormapsRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 83;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::ListInstalledColormapsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::ListInstalledColormapsReply>
detail::ReadReply<XProto::ListInstalledColormapsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::ListInstalledColormapsReply>();

  auto& cmaps_len = (*reply).cmaps_len;
  auto& cmaps = (*reply).cmaps;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // cmaps_len
  Read(&cmaps_len, &buf);

  // pad1
  Pad(&buf, 22);

  // cmaps
  cmaps.resize(cmaps_len);
  for (auto& cmaps_elem : cmaps) {
    Read(&cmaps_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::AllocColorReply>
XProto::AllocColor(
    const XProto::AllocColorRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& red = request.red;
  auto& green = request.green;
  auto& blue = request.blue;

  // major_opcode
  uint8_t major_opcode = 84;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // red
  Write(&red, &buf);

  // green
  Write(&green, &buf);

  // blue
  Write(&blue, &buf);

  // pad1
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<XProto::AllocColorReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::AllocColorReply>
detail::ReadReply<XProto::AllocColorReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::AllocColorReply>();

  auto& red = (*reply).red;
  auto& green = (*reply).green;
  auto& blue = (*reply).blue;
  auto& pixel = (*reply).pixel;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // red
  Read(&red, &buf);

  // green
  Read(&green, &buf);

  // blue
  Read(&blue, &buf);

  // pad1
  Pad(&buf, 2);

  // pixel
  Read(&pixel, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::AllocNamedColorReply>
XProto::AllocNamedColor(
    const XProto::AllocNamedColorRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;
  uint16_t name_len;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 85;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    Write(&name_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::AllocNamedColorReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::AllocNamedColorReply>
detail::ReadReply<XProto::AllocNamedColorReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::AllocNamedColorReply>();

  auto& pixel = (*reply).pixel;
  auto& exact_red = (*reply).exact_red;
  auto& exact_green = (*reply).exact_green;
  auto& exact_blue = (*reply).exact_blue;
  auto& visual_red = (*reply).visual_red;
  auto& visual_green = (*reply).visual_green;
  auto& visual_blue = (*reply).visual_blue;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pixel
  Read(&pixel, &buf);

  // exact_red
  Read(&exact_red, &buf);

  // exact_green
  Read(&exact_green, &buf);

  // exact_blue
  Read(&exact_blue, &buf);

  // visual_red
  Read(&visual_red, &buf);

  // visual_green
  Read(&visual_green, &buf);

  // visual_blue
  Read(&visual_blue, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::AllocColorCellsReply>
XProto::AllocColorCells(
    const XProto::AllocColorCellsRequest& request) {
  WriteBuffer buf;

  auto& contiguous = request.contiguous;
  auto& cmap = request.cmap;
  auto& colors = request.colors;
  auto& planes = request.planes;

  // major_opcode
  uint8_t major_opcode = 86;
  Write(&major_opcode, &buf);

  // contiguous
  Write(&contiguous, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // colors
  Write(&colors, &buf);

  // planes
  Write(&planes, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::AllocColorCellsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::AllocColorCellsReply>
detail::ReadReply<XProto::AllocColorCellsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::AllocColorCellsReply>();

  uint16_t pixels_len;
  uint16_t masks_len;
  auto& pixels = (*reply).pixels;
  auto& masks = (*reply).masks;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pixels_len
  Read(&pixels_len, &buf);

  // masks_len
  Read(&masks_len, &buf);

  // pad1
  Pad(&buf, 20);

  // pixels
  pixels.resize(pixels_len);
  for (auto& pixels_elem : pixels) {
    Read(&pixels_elem, &buf);
  }

  // masks
  masks.resize(masks_len);
  for (auto& masks_elem : masks) {
    Read(&masks_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::AllocColorPlanesReply>
XProto::AllocColorPlanes(
    const XProto::AllocColorPlanesRequest& request) {
  WriteBuffer buf;

  auto& contiguous = request.contiguous;
  auto& cmap = request.cmap;
  auto& colors = request.colors;
  auto& reds = request.reds;
  auto& greens = request.greens;
  auto& blues = request.blues;

  // major_opcode
  uint8_t major_opcode = 87;
  Write(&major_opcode, &buf);

  // contiguous
  Write(&contiguous, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // colors
  Write(&colors, &buf);

  // reds
  Write(&reds, &buf);

  // greens
  Write(&greens, &buf);

  // blues
  Write(&blues, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::AllocColorPlanesReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::AllocColorPlanesReply>
detail::ReadReply<XProto::AllocColorPlanesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::AllocColorPlanesReply>();

  uint16_t pixels_len;
  auto& red_mask = (*reply).red_mask;
  auto& green_mask = (*reply).green_mask;
  auto& blue_mask = (*reply).blue_mask;
  auto& pixels = (*reply).pixels;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pixels_len
  Read(&pixels_len, &buf);

  // pad1
  Pad(&buf, 2);

  // red_mask
  Read(&red_mask, &buf);

  // green_mask
  Read(&green_mask, &buf);

  // blue_mask
  Read(&blue_mask, &buf);

  // pad2
  Pad(&buf, 8);

  // pixels
  pixels.resize(pixels_len);
  for (auto& pixels_elem : pixels) {
    Read(&pixels_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::FreeColors(
    const XProto::FreeColorsRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& plane_mask = request.plane_mask;
  auto& pixels = request.pixels;
  size_t pixels_len = pixels.size();

  // major_opcode
  uint8_t major_opcode = 88;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // plane_mask
  Write(&plane_mask, &buf);

  // pixels
  DCHECK_EQ(static_cast<size_t>(pixels_len), pixels.size());
  for (auto& pixels_elem : pixels) {
    Write(&pixels_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::StoreColors(
    const XProto::StoreColorsRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& items = request.items;
  size_t items_len = items.size();

  // major_opcode
  uint8_t major_opcode = 89;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // items
  DCHECK_EQ(static_cast<size_t>(items_len), items.size());
  for (auto& items_elem : items) {
    auto& pixel = items_elem.pixel;
    auto& red = items_elem.red;
    auto& green = items_elem.green;
    auto& blue = items_elem.blue;
    auto& flags = items_elem.flags;

    // pixel
    Write(&pixel, &buf);

    // red
    Write(&red, &buf);

    // green
    Write(&green, &buf);

    // blue
    Write(&blue, &buf);

    // flags
    uint8_t tmp80;
    tmp80 = static_cast<uint8_t>(flags);
    Write(&tmp80, &buf);

    // pad0
    Pad(&buf, 1);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::StoreNamedColor(
    const XProto::StoreNamedColorRequest& request) {
  WriteBuffer buf;

  auto& flags = request.flags;
  auto& cmap = request.cmap;
  auto& pixel = request.pixel;
  uint16_t name_len;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 90;
  Write(&major_opcode, &buf);

  // flags
  uint8_t tmp81;
  tmp81 = static_cast<uint8_t>(flags);
  Write(&tmp81, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // pixel
  Write(&pixel, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad0
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    Write(&name_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::QueryColorsReply>
XProto::QueryColors(
    const XProto::QueryColorsRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& pixels = request.pixels;
  size_t pixels_len = pixels.size();

  // major_opcode
  uint8_t major_opcode = 91;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // pixels
  DCHECK_EQ(static_cast<size_t>(pixels_len), pixels.size());
  for (auto& pixels_elem : pixels) {
    Write(&pixels_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryColorsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryColorsReply>
detail::ReadReply<XProto::QueryColorsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryColorsReply>();

  auto& colors_len = (*reply).colors_len;
  auto& colors = (*reply).colors;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // colors_len
  Read(&colors_len, &buf);

  // pad1
  Pad(&buf, 22);

  // colors
  colors.resize(colors_len);
  for (auto& colors_elem : colors) {
    auto& red = colors_elem.red;
    auto& green = colors_elem.green;
    auto& blue = colors_elem.blue;

    // red
    Read(&red, &buf);

    // green
    Read(&green, &buf);

    // blue
    Read(&blue, &buf);

    // pad0
    Pad(&buf, 2);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::LookupColorReply>
XProto::LookupColor(
    const XProto::LookupColorRequest& request) {
  WriteBuffer buf;

  auto& cmap = request.cmap;
  uint16_t name_len;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 92;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    Write(&name_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::LookupColorReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::LookupColorReply>
detail::ReadReply<XProto::LookupColorReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::LookupColorReply>();

  auto& exact_red = (*reply).exact_red;
  auto& exact_green = (*reply).exact_green;
  auto& exact_blue = (*reply).exact_blue;
  auto& visual_red = (*reply).visual_red;
  auto& visual_green = (*reply).visual_green;
  auto& visual_blue = (*reply).visual_blue;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // exact_red
  Read(&exact_red, &buf);

  // exact_green
  Read(&exact_green, &buf);

  // exact_blue
  Read(&exact_blue, &buf);

  // visual_red
  Read(&visual_red, &buf);

  // visual_green
  Read(&visual_green, &buf);

  // visual_blue
  Read(&visual_blue, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::CreateCursor(
    const XProto::CreateCursorRequest& request) {
  WriteBuffer buf;

  auto& cid = request.cid;
  auto& source = request.source;
  auto& mask = request.mask;
  auto& fore_red = request.fore_red;
  auto& fore_green = request.fore_green;
  auto& fore_blue = request.fore_blue;
  auto& back_red = request.back_red;
  auto& back_green = request.back_green;
  auto& back_blue = request.back_blue;
  auto& x = request.x;
  auto& y = request.y;

  // major_opcode
  uint8_t major_opcode = 93;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // source
  Write(&source, &buf);

  // mask
  Write(&mask, &buf);

  // fore_red
  Write(&fore_red, &buf);

  // fore_green
  Write(&fore_green, &buf);

  // fore_blue
  Write(&fore_blue, &buf);

  // back_red
  Write(&back_red, &buf);

  // back_green
  Write(&back_green, &buf);

  // back_blue
  Write(&back_blue, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::CreateGlyphCursor(
    const XProto::CreateGlyphCursorRequest& request) {
  WriteBuffer buf;

  auto& cid = request.cid;
  auto& source_font = request.source_font;
  auto& mask_font = request.mask_font;
  auto& source_char = request.source_char;
  auto& mask_char = request.mask_char;
  auto& fore_red = request.fore_red;
  auto& fore_green = request.fore_green;
  auto& fore_blue = request.fore_blue;
  auto& back_red = request.back_red;
  auto& back_green = request.back_green;
  auto& back_blue = request.back_blue;

  // major_opcode
  uint8_t major_opcode = 94;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // source_font
  Write(&source_font, &buf);

  // mask_font
  Write(&mask_font, &buf);

  // source_char
  Write(&source_char, &buf);

  // mask_char
  Write(&mask_char, &buf);

  // fore_red
  Write(&fore_red, &buf);

  // fore_green
  Write(&fore_green, &buf);

  // fore_blue
  Write(&fore_blue, &buf);

  // back_red
  Write(&back_red, &buf);

  // back_green
  Write(&back_green, &buf);

  // back_blue
  Write(&back_blue, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::FreeCursor(
    const XProto::FreeCursorRequest& request) {
  WriteBuffer buf;

  auto& cursor = request.cursor;

  // major_opcode
  uint8_t major_opcode = 95;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cursor
  Write(&cursor, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::RecolorCursor(
    const XProto::RecolorCursorRequest& request) {
  WriteBuffer buf;

  auto& cursor = request.cursor;
  auto& fore_red = request.fore_red;
  auto& fore_green = request.fore_green;
  auto& fore_blue = request.fore_blue;
  auto& back_red = request.back_red;
  auto& back_green = request.back_green;
  auto& back_blue = request.back_blue;

  // major_opcode
  uint8_t major_opcode = 96;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cursor
  Write(&cursor, &buf);

  // fore_red
  Write(&fore_red, &buf);

  // fore_green
  Write(&fore_green, &buf);

  // fore_blue
  Write(&fore_blue, &buf);

  // back_red
  Write(&back_red, &buf);

  // back_green
  Write(&back_green, &buf);

  // back_blue
  Write(&back_blue, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::QueryBestSizeReply>
XProto::QueryBestSize(
    const XProto::QueryBestSizeRequest& request) {
  WriteBuffer buf;

  auto& c_class = request.c_class;
  auto& drawable = request.drawable;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 97;
  Write(&major_opcode, &buf);

  // c_class
  uint8_t tmp82;
  tmp82 = static_cast<uint8_t>(c_class);
  Write(&tmp82, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryBestSizeReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryBestSizeReply>
detail::ReadReply<XProto::QueryBestSizeReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryBestSizeReply>();

  auto& width = (*reply).width;
  auto& height = (*reply).height;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::QueryExtensionReply>
XProto::QueryExtension(
    const XProto::QueryExtensionRequest& request) {
  WriteBuffer buf;

  uint16_t name_len;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 98;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    Write(&name_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::QueryExtensionReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::QueryExtensionReply>
detail::ReadReply<XProto::QueryExtensionReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::QueryExtensionReply>();

  auto& present = (*reply).present;
  auto& major_opcode = (*reply).major_opcode;
  auto& first_event = (*reply).first_event;
  auto& first_error = (*reply).first_error;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // present
  Read(&present, &buf);

  // major_opcode
  Read(&major_opcode, &buf);

  // first_event
  Read(&first_event, &buf);

  // first_error
  Read(&first_error, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::ListExtensionsReply>
XProto::ListExtensions(
    const XProto::ListExtensionsRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 99;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::ListExtensionsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::ListExtensionsReply>
detail::ReadReply<XProto::ListExtensionsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::ListExtensionsReply>();

  uint8_t names_len;
  auto& names = (*reply).names;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // names_len
  Read(&names_len, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // names
  names.resize(names_len);
  for (auto& names_elem : names) {
    uint8_t name_len;
    auto& name = names_elem.name;

    // name_len
    Read(&name_len, &buf);

    // name
    name.resize(name_len);
    for (auto& name_elem : name) {
      Read(&name_elem, &buf);
    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeKeyboardMapping(
    const XProto::ChangeKeyboardMappingRequest& request) {
  WriteBuffer buf;

  auto& keycode_count = request.keycode_count;
  auto& first_keycode = request.first_keycode;
  auto& keysyms_per_keycode = request.keysyms_per_keycode;
  auto& keysyms = request.keysyms;
  size_t keysyms_len = keysyms.size();

  // major_opcode
  uint8_t major_opcode = 100;
  Write(&major_opcode, &buf);

  // keycode_count
  Write(&keycode_count, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // first_keycode
  Write(&first_keycode, &buf);

  // keysyms_per_keycode
  Write(&keysyms_per_keycode, &buf);

  // pad0
  Pad(&buf, 2);

  // keysyms
  DCHECK_EQ(static_cast<size_t>((keycode_count) * (keysyms_per_keycode)), keysyms.size());
  for (auto& keysyms_elem : keysyms) {
    Write(&keysyms_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetKeyboardMappingReply>
XProto::GetKeyboardMapping(
    const XProto::GetKeyboardMappingRequest& request) {
  WriteBuffer buf;

  auto& first_keycode = request.first_keycode;
  auto& count = request.count;

  // major_opcode
  uint8_t major_opcode = 101;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // first_keycode
  Write(&first_keycode, &buf);

  // count
  Write(&count, &buf);

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetKeyboardMappingReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetKeyboardMappingReply>
detail::ReadReply<XProto::GetKeyboardMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetKeyboardMappingReply>();

  auto& keysyms_per_keycode = (*reply).keysyms_per_keycode;
  auto& keysyms = (*reply).keysyms;
  size_t keysyms_len = keysyms.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // keysyms_per_keycode
  Read(&keysyms_per_keycode, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // keysyms
  keysyms.resize(length);
  for (auto& keysyms_elem : keysyms) {
    Read(&keysyms_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeKeyboardControl(
    const XProto::ChangeKeyboardControlRequest& request) {
  WriteBuffer buf;

  auto& value_mask = request.value_mask;
  auto& value_list = request.value_list;

  // major_opcode
  uint8_t major_opcode = 102;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // value_mask
  uint32_t tmp83;
  tmp83 = static_cast<uint32_t>(value_mask);
  Write(&tmp83, &buf);

  // value_list
  {
    auto& key_click_percent = value_list.key_click_percent;
    auto& bell_percent = value_list.bell_percent;
    auto& bell_pitch = value_list.bell_pitch;
    auto& bell_duration = value_list.bell_duration;
    auto& led = value_list.led;
    auto& led_mode = value_list.led_mode;
    auto& key = value_list.key;
    auto& auto_repeat_mode = value_list.auto_repeat_mode;

    auto value_list_expr = value_mask;
    if (BitAnd(value_list_expr, Keyboard::KeyClickPercent)) {
      // key_click_percent
      Write(&key_click_percent, &buf);
    }
    if (BitAnd(value_list_expr, Keyboard::BellPercent)) {
      // bell_percent
      Write(&bell_percent, &buf);
    }
    if (BitAnd(value_list_expr, Keyboard::BellPitch)) {
      // bell_pitch
      Write(&bell_pitch, &buf);
    }
    if (BitAnd(value_list_expr, Keyboard::BellDuration)) {
      // bell_duration
      Write(&bell_duration, &buf);
    }
    if (BitAnd(value_list_expr, Keyboard::Led)) {
      // led
      Write(&led, &buf);
    }
    if (BitAnd(value_list_expr, Keyboard::LedMode)) {
      // led_mode
      uint32_t tmp84;
      tmp84 = static_cast<uint32_t>(led_mode);
      Write(&tmp84, &buf);
    }
    if (BitAnd(value_list_expr, Keyboard::Key)) {
      // key
      Write(&key, &buf);
    }
    if (BitAnd(value_list_expr, Keyboard::AutoRepeatMode)) {
      // auto_repeat_mode
      uint32_t tmp85;
      tmp85 = static_cast<uint32_t>(auto_repeat_mode);
      Write(&tmp85, &buf);
    }
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetKeyboardControlReply>
XProto::GetKeyboardControl(
    const XProto::GetKeyboardControlRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 103;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetKeyboardControlReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetKeyboardControlReply>
detail::ReadReply<XProto::GetKeyboardControlReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetKeyboardControlReply>();

  auto& global_auto_repeat = (*reply).global_auto_repeat;
  auto& led_mask = (*reply).led_mask;
  auto& key_click_percent = (*reply).key_click_percent;
  auto& bell_percent = (*reply).bell_percent;
  auto& bell_pitch = (*reply).bell_pitch;
  auto& bell_duration = (*reply).bell_duration;
  auto& auto_repeats = (*reply).auto_repeats;
  size_t auto_repeats_len = auto_repeats.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // global_auto_repeat
  uint8_t tmp86;
  Read(&tmp86, &buf);
  global_auto_repeat = static_cast<XProto::AutoRepeatMode>(tmp86);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // led_mask
  Read(&led_mask, &buf);

  // key_click_percent
  Read(&key_click_percent, &buf);

  // bell_percent
  Read(&bell_percent, &buf);

  // bell_pitch
  Read(&bell_pitch, &buf);

  // bell_duration
  Read(&bell_duration, &buf);

  // pad0
  Pad(&buf, 2);

  // auto_repeats
  for (auto& auto_repeats_elem : auto_repeats) {
    Read(&auto_repeats_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::Bell(
    const XProto::BellRequest& request) {
  WriteBuffer buf;

  auto& percent = request.percent;

  // major_opcode
  uint8_t major_opcode = 104;
  Write(&major_opcode, &buf);

  // percent
  Write(&percent, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ChangePointerControl(
    const XProto::ChangePointerControlRequest& request) {
  WriteBuffer buf;

  auto& acceleration_numerator = request.acceleration_numerator;
  auto& acceleration_denominator = request.acceleration_denominator;
  auto& threshold = request.threshold;
  auto& do_acceleration = request.do_acceleration;
  auto& do_threshold = request.do_threshold;

  // major_opcode
  uint8_t major_opcode = 105;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // acceleration_numerator
  Write(&acceleration_numerator, &buf);

  // acceleration_denominator
  Write(&acceleration_denominator, &buf);

  // threshold
  Write(&threshold, &buf);

  // do_acceleration
  Write(&do_acceleration, &buf);

  // do_threshold
  Write(&do_threshold, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetPointerControlReply>
XProto::GetPointerControl(
    const XProto::GetPointerControlRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 106;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetPointerControlReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetPointerControlReply>
detail::ReadReply<XProto::GetPointerControlReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetPointerControlReply>();

  auto& acceleration_numerator = (*reply).acceleration_numerator;
  auto& acceleration_denominator = (*reply).acceleration_denominator;
  auto& threshold = (*reply).threshold;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // acceleration_numerator
  Read(&acceleration_numerator, &buf);

  // acceleration_denominator
  Read(&acceleration_denominator, &buf);

  // threshold
  Read(&threshold, &buf);

  // pad1
  Pad(&buf, 18);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetScreenSaver(
    const XProto::SetScreenSaverRequest& request) {
  WriteBuffer buf;

  auto& timeout = request.timeout;
  auto& interval = request.interval;
  auto& prefer_blanking = request.prefer_blanking;
  auto& allow_exposures = request.allow_exposures;

  // major_opcode
  uint8_t major_opcode = 107;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // timeout
  Write(&timeout, &buf);

  // interval
  Write(&interval, &buf);

  // prefer_blanking
  uint8_t tmp87;
  tmp87 = static_cast<uint8_t>(prefer_blanking);
  Write(&tmp87, &buf);

  // allow_exposures
  uint8_t tmp88;
  tmp88 = static_cast<uint8_t>(allow_exposures);
  Write(&tmp88, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::GetScreenSaverReply>
XProto::GetScreenSaver(
    const XProto::GetScreenSaverRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 108;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetScreenSaverReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetScreenSaverReply>
detail::ReadReply<XProto::GetScreenSaverReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetScreenSaverReply>();

  auto& timeout = (*reply).timeout;
  auto& interval = (*reply).interval;
  auto& prefer_blanking = (*reply).prefer_blanking;
  auto& allow_exposures = (*reply).allow_exposures;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // timeout
  Read(&timeout, &buf);

  // interval
  Read(&interval, &buf);

  // prefer_blanking
  uint8_t tmp89;
  Read(&tmp89, &buf);
  prefer_blanking = static_cast<XProto::Blanking>(tmp89);

  // allow_exposures
  uint8_t tmp90;
  Read(&tmp90, &buf);
  allow_exposures = static_cast<XProto::Exposures>(tmp90);

  // pad1
  Pad(&buf, 18);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeHosts(
    const XProto::ChangeHostsRequest& request) {
  WriteBuffer buf;

  auto& mode = request.mode;
  auto& family = request.family;
  auto& address_len = request.address_len;
  auto& address = request.address;

  // major_opcode
  uint8_t major_opcode = 109;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp91;
  tmp91 = static_cast<uint8_t>(mode);
  Write(&tmp91, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // family
  uint8_t tmp92;
  tmp92 = static_cast<uint8_t>(family);
  Write(&tmp92, &buf);

  // pad0
  Pad(&buf, 1);

  // address_len
  Write(&address_len, &buf);

  // address
  DCHECK_EQ(static_cast<size_t>(address_len), address.size());
  for (auto& address_elem : address) {
    Write(&address_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::ListHostsReply>
XProto::ListHosts(
    const XProto::ListHostsRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 110;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::ListHostsReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::ListHostsReply>
detail::ReadReply<XProto::ListHostsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::ListHostsReply>();

  auto& mode = (*reply).mode;
  uint16_t hosts_len;
  auto& hosts = (*reply).hosts;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // mode
  uint8_t tmp93;
  Read(&tmp93, &buf);
  mode = static_cast<XProto::AccessControl>(tmp93);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // hosts_len
  Read(&hosts_len, &buf);

  // pad0
  Pad(&buf, 22);

  // hosts
  hosts.resize(hosts_len);
  for (auto& hosts_elem : hosts) {
    auto& family = hosts_elem.family;
    auto& address_len = hosts_elem.address_len;
    auto& address = hosts_elem.address;

    // family
    uint8_t tmp94;
    Read(&tmp94, &buf);
    family = static_cast<XProto::Family>(tmp94);

    // pad0
    Pad(&buf, 1);

    // address_len
    Read(&address_len, &buf);

    // address
    address.resize(address_len);
    for (auto& address_elem : address) {
      Read(&address_elem, &buf);
    }

    // pad1
    Align(&buf, 4);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetAccessControl(
    const XProto::SetAccessControlRequest& request) {
  WriteBuffer buf;

  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = 111;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp95;
  tmp95 = static_cast<uint8_t>(mode);
  Write(&tmp95, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::SetCloseDownMode(
    const XProto::SetCloseDownModeRequest& request) {
  WriteBuffer buf;

  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = 112;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp96;
  tmp96 = static_cast<uint8_t>(mode);
  Write(&tmp96, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::KillClient(
    const XProto::KillClientRequest& request) {
  WriteBuffer buf;

  auto& resource = request.resource;

  // major_opcode
  uint8_t major_opcode = 113;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // resource
  Write(&resource, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::RotateProperties(
    const XProto::RotatePropertiesRequest& request) {
  WriteBuffer buf;

  auto& window = request.window;
  auto& atoms_len = request.atoms_len;
  auto& delta = request.delta;
  auto& atoms = request.atoms;

  // major_opcode
  uint8_t major_opcode = 114;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // atoms_len
  Write(&atoms_len, &buf);

  // delta
  Write(&delta, &buf);

  // atoms
  DCHECK_EQ(static_cast<size_t>(atoms_len), atoms.size());
  for (auto& atoms_elem : atoms) {
    Write(&atoms_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<void>
XProto::ForceScreenSaver(
    const XProto::ForceScreenSaverRequest& request) {
  WriteBuffer buf;

  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = 115;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp97;
  tmp97 = static_cast<uint8_t>(mode);
  Write(&tmp97, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

Future<XProto::SetPointerMappingReply>
XProto::SetPointerMapping(
    const XProto::SetPointerMappingRequest& request) {
  WriteBuffer buf;

  uint8_t map_len;
  auto& map = request.map;

  // major_opcode
  uint8_t major_opcode = 116;
  Write(&major_opcode, &buf);

  // map_len
  map_len = map.size();
  Write(&map_len, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // map
  DCHECK_EQ(static_cast<size_t>(map_len), map.size());
  for (auto& map_elem : map) {
    Write(&map_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::SetPointerMappingReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::SetPointerMappingReply>
detail::ReadReply<XProto::SetPointerMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::SetPointerMappingReply>();

  auto& status = (*reply).status;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp98;
  Read(&tmp98, &buf);
  status = static_cast<XProto::MappingStatus>(tmp98);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::GetPointerMappingReply>
XProto::GetPointerMapping(
    const XProto::GetPointerMappingRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 117;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetPointerMappingReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetPointerMappingReply>
detail::ReadReply<XProto::GetPointerMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetPointerMappingReply>();

  auto& map_len = (*reply).map_len;
  auto& map = (*reply).map;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // map_len
  Read(&map_len, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // map
  map.resize(map_len);
  for (auto& map_elem : map) {
    Read(&map_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::SetModifierMappingReply>
XProto::SetModifierMapping(
    const XProto::SetModifierMappingRequest& request) {
  WriteBuffer buf;

  auto& keycodes_per_modifier = request.keycodes_per_modifier;
  auto& keycodes = request.keycodes;
  size_t keycodes_len = keycodes.size();

  // major_opcode
  uint8_t major_opcode = 118;
  Write(&major_opcode, &buf);

  // keycodes_per_modifier
  Write(&keycodes_per_modifier, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // keycodes
  DCHECK_EQ(static_cast<size_t>((keycodes_per_modifier) * (8)), keycodes.size());
  for (auto& keycodes_elem : keycodes) {
    Write(&keycodes_elem, &buf);
  }

  Align(&buf, 4);

  return x11::SendRequest<XProto::SetModifierMappingReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::SetModifierMappingReply>
detail::ReadReply<XProto::SetModifierMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::SetModifierMappingReply>();

  auto& status = (*reply).status;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp99;
  Read(&tmp99, &buf);
  status = static_cast<XProto::MappingStatus>(tmp99);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<XProto::GetModifierMappingReply>
XProto::GetModifierMapping(
    const XProto::GetModifierMappingRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 119;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<XProto::GetModifierMappingReply>(display_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<XProto::GetModifierMappingReply>
detail::ReadReply<XProto::GetModifierMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<XProto::GetModifierMappingReply>();

  auto& keycodes_per_modifier = (*reply).keycodes_per_modifier;
  auto& keycodes = (*reply).keycodes;
  size_t keycodes_len = keycodes.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // keycodes_per_modifier
  Read(&keycodes_per_modifier, &buf);

  // sequence
  uint16_t sequence;
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // keycodes
  keycodes.resize((keycodes_per_modifier) * (8));
  for (auto& keycodes_elem : keycodes) {
    Read(&keycodes_elem, &buf);
  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::NoOperation(
    const XProto::NoOperationRequest& request) {
  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 127;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(display_, &buf);
}

}  // namespace x11
